(*

  application: image filtering
  state: experimental
  revision: $Id: filt_config.ml,v 1.10 2004/11/20 14:25:35 tobias Exp $
 
  Metaprogramming Project
  Authors: Christoph A. Herrmann and Tobias Langhammer
  University of Passau, Department of Informatics, Chair of Programming

 Copyright (C) 2004    University of Passau, Department of Informatics
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*)

open Printf
open Ppm

type compilation = (int * int) -> int
type config = { 
  mutable ic:in_channel;
  mutable oc:out_channel;
  mutable osuffix:string;
  mutable iter_indep: bool;
  mutable staged:bool;
  mutable simplify:bool;
  mutable compiler: Ppm.ppm_image -> int -> Defs.exp -> compilation;
  mutable time:bool;
  mutable iter:int;
  mutable nostage_iter:int;
  mutable timings_iter:int;
  mutable norm_range:bool;
  mutable benchmarks:bool;
  mutable margintop:int;
  mutable marginbottom:int;
  mutable marginleft:int;
  mutable marginright:int;
  mutable verbose:int;
  mutable red_formula:string;
  mutable green_formula:string;
  mutable blue_formula:string;
}

(* postprocessing after command line evaluation *)
let postprocess cfg = begin
  let staged img iter exp = 
      let exp' = Staticeval.simplify exp img.width img.height iter in
      Stagedinter.compile exp' img 
  in
  let unstaged img iter exp = 
      let exp' = Staticeval.simplify exp img.width img.height iter in
      Unstagedinter.compile exp' img 
  in
  let staged_no_simplify img iter exp = 
      Stagedinter.compile exp img 
  in
  let unstaged_no_simplify img iter exp = 
      Unstagedinter.compile exp img 
  in
  cfg.compiler <- 
    begin match (cfg.staged, cfg.simplify) with
    | (true, true) -> staged
    | (true, false) -> staged_no_simplify
    | (false, true) -> unstaged
    | (false, false) -> unstaged_no_simplify
    end;
  cfg
end

let get () = (* unit -> config *)
  begin
    let cfg = {
        ic = stdin;
        oc = stdout;
        osuffix = "";
        iter = 1;
        nostage_iter = 1;
        timings_iter = 1;
        iter_indep = true;
        staged = true;
        simplify = true;
        compiler = (fun _ _ _ _ -> 4711);
        time = false;
        norm_range = true;
        benchmarks = false;
        margintop = 0;
        marginleft = 0;
        marginright = 0;
        marginbottom = 0;
        verbose = 0;
        red_formula = "red(row,col);";
        green_formula = "green(row,col);";
        blue_formula = "blue(row,col);"
      } 
    in
    let print_help () = 
      begin
        prerr_string "usage: ";
        prerr_string Sys.executable_name;
        prerr_string " -e <expr> [option ...]\n";
        prerr_string "\nwhere\n";
        prerr_string " (-red|-green|-blue) <expr>\n";
        prerr_string "                       set formula for single channel\n";
        prerr_string "  -gray <expr>         set formula for all channels\n";
        prerr_string "  -f <file>            file with 3 formulae for \n";
        prerr_string "                       red, green and blue channel.\n";
        prerr_string "                       Each formula has to be in a separate line.\n";
        prerr_string "  -i <img>             input ppm image (default: stdin)\n";
        prerr_string "  -o <img>             output ppm image (default: stdout)\n";
        prerr_string "  -m <size>            all margins\n";
        prerr_string "  -mt <size>           margin top\n";
        prerr_string "  -mb <size>           margin bottom\n";
        prerr_string "  -ml <size>           margin left\n";
        prerr_string "  -mr <size>           margin right\n";
        prerr_string "  -v                   increase verbosity\n";
        prerr_string "  -n <n>               number of iterations (default: 1)\n";
        prerr_string "  -ns <suffix>         write intermediate files of iteration to <n><suffix>\n";
        prerr_string "  -ni                  independent filter on original input (default)\n";
        prerr_string "  -nc                  consecutive filter on previous iteration\n";
        prerr_string "\n benchmark flags: \n";
        prerr_string "  -staged              staging optimization (default)\n";
        prerr_string "  -unstaged            staging optimization off\n";
        prerr_string "  -simplify            expression simplification optimization (default)\n";
        prerr_string "  -nosimplify          expression simplification optimization off\n";
        prerr_string "\n";
        prerr_string "  -nostage_n <n>       single staging and <n> iterations\n";
        prerr_string "  -timings_n <n>       number of timings to be done\n";
        prerr_string "  -time                output time of run\n";
        prerr_string "\n  -h                   this help\n";
        exit 0;
      end in
    let (i:int ref) = ref 0 in
    while !i <> Array.length Sys.argv-1 do
      i := !i + 1;
      match Sys.argv.(!i) with
        "-gray" -> 
          begin
            i := !i + 1;
            cfg.red_formula <- Sys.argv.(!i)^";";
            cfg.green_formula <- Sys.argv.(!i)^";";
            cfg.blue_formula <- Sys.argv.(!i)^";";
          end
      | "-red" -> 
          begin
            i := !i + 1;
            cfg.red_formula <- Sys.argv.(!i)^";"
          end
      | "-green" -> 
          begin
            i := !i + 1;
            cfg.green_formula <- Sys.argv.(!i)^";"
          end
      | "-blue" -> 
          begin
            i := !i + 1;
            cfg.blue_formula <- Sys.argv.(!i)^";"
          end
      | "-f" -> 
          begin
            i := !i + 1;
            let ic = open_in Sys.argv.(!i) in
            let ins = ref (input_line ic) in
            try while true  do
              ins := !ins ^ " " ^ input_line ic
            done with End_of_file -> ();
            close_in ic;
            cfg.red_formula <- !ins;
            cfg.green_formula <- "";
            cfg.blue_formula <- "";
          end
      | "-i" -> 
          begin
            i := !i + 1;
            cfg.ic <- open_in Sys.argv.(!i)
          end
      | "-o" -> 
          begin
            i := !i + 1;
            cfg.oc <- open_out Sys.argv.(!i)
          end
      | "-mt" -> 
          begin
            i := !i + 1;
            cfg.margintop <- int_of_string Sys.argv.(!i)
          end
      | "-mb" -> 
          begin
            i := !i + 1;
            cfg.marginbottom <- int_of_string Sys.argv.(!i)
          end
      | "-ml" -> 
          begin
            i := !i + 1;
            cfg.marginleft <- int_of_string Sys.argv.(!i)
          end
      | "-mr" -> 
          begin
            i := !i + 1;
            cfg.marginright <- int_of_string Sys.argv.(!i)
          end
      | "-m" -> 
          begin
            i := !i + 1;
            cfg.margintop    <- int_of_string Sys.argv.(!i);
            cfg.marginleft   <- int_of_string Sys.argv.(!i);
            cfg.marginright  <- int_of_string Sys.argv.(!i);
            cfg.marginbottom <- int_of_string Sys.argv.(!i);
          end
      | "-n" -> 
          begin
            i := !i + 1;
            cfg.iter  <- int_of_string Sys.argv.(!i)
          end
      | "-nostage_n" -> 
          begin
            i := !i + 1;
            cfg.nostage_iter  <- int_of_string Sys.argv.(!i)
          end
      | "-timings_n" -> 
          begin
            i := !i + 1;
            cfg.timings_iter  <- int_of_string Sys.argv.(!i)
          end
      | "-v" -> 
          begin
            cfg.verbose  <- 1;
          end
      | "-ns" ->
          begin
            i := !i + 1;
            cfg.osuffix <- Sys.argv.(!i)
          end
      | "-nc" -> cfg.iter_indep <- false
      | "-ni" -> cfg.iter_indep <- true
      | "-time" -> cfg.time <- true
      | "-staged" -> cfg.staged <- true
      | "-unstaged" -> cfg.staged <- false
      | "-simplify" -> cfg.simplify <- true
      | "-nosimplify" -> cfg.simplify <- false
      | "-bench" -> cfg.benchmarks <- true
      | "-h" -> print_help ()
      | "--help" -> print_help ()
      | unknown -> 
          begin 
            prerr_string "invalid parameter '";
            prerr_string unknown;
            prerr_string "'\n";
            exit 0;
          end
    done;
    postprocess cfg
  end

let finalize cfg = 
  begin
    if cfg.ic <> stdin then close_in cfg.ic;
    if cfg.oc <> stdout then close_out cfg.oc;
  end   

