(*

  application: image filtering
  state: experimental
  revision: $Id: main.ml,v 1.26 2004/11/20 14:25:35 tobias Exp $
 
  Metaprogramming Project
  Authors: Christoph A. Herrmann and Tobias Langhammer
  University of Passau, Department of Informatics, Chair of Programming

 Copyright (C) 2004    University of Passau, Department of Informatics
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*)


open Printf
open Filt_config
open Ppm 

let exec_times n f =
  for i = 1 to n do
    f ()
  done

 
let parse formulae = begin
  let lexbuf = Lexing.from_string formulae in
  let res = ref [] in
  begin try
    while true do
      res := 
        Filt_parser.main Filt_lexer.token lexbuf::!res
    done
  with 
  | Parsing.Parse_error ->
    begin
      let pos = lexbuf.Lexing.lex_start_p.Lexing.pos_cnum in
      prerr_string "Parse error in the following expression:\n\n";
      prerr_string lexbuf.Lexing.lex_buffer;
      prerr_string "\nUnexpected character '";
      prerr_char lexbuf.Lexing.lex_buffer.[pos];
      prerr_string "' at position ";
      prerr_int pos;
      prerr_string "!\n";
      prerr_string "\nexiting.\n";
      flush stderr;
      raise Parsing.Parse_error; 
    end
  | Filt_lexer.Eof -> ();
  end;
  List.rev !res
end
  
  
let runfilter cfg in_img out_img =
  begin
    let formulae = cfg.red_formula ^ cfg.green_formula ^ cfg.blue_formula in
    let exps = parse formulae in
    let run () = begin
      for i = 1 to cfg.iter do
        let filters = List.map (cfg.compiler !in_img i) exps in
        for j = 1 to cfg.nostage_iter do (* this is to get the staging overhead *)
          let top = cfg.margintop in
          let bottom = (!in_img).height-1-cfg.marginbottom in
          let left = cfg.marginleft in
          let right = (!in_img).width-1-cfg.marginright in
          for row = top to bottom do
            for col = left to right do
              let colors =  List.map (fun f -> f (row,col)) filters in
              Ppm.set_point !out_img row col 
                (Ppm.point_of_array (Array.of_list colors))
            done
          done
        done;
        if cfg.osuffix <> "" then begin
            let oc = open_out (string_of_int i ^cfg.osuffix) in
            Ppm.output_ppm oc !out_img;
            close_out oc;
        end;
        if not cfg.iter_indep then begin
          let tmp = !out_img in
          out_img := !in_img;
          in_img := tmp;
        end;
      done;
      if cfg.iter_indep then begin
        let tmp = !out_img in
        out_img := !in_img;
        in_img := tmp;
      end
    end in
    let label = 
      string_of_int (!out_img).width ^ ";" 
      ^ string_of_int (!out_img).height 
      ^ ";byte;" ^ 
      (if cfg.staged then "staged;" else "inter;") ^
      if cfg.simplify then "sfy" else "noSfy" in
    Timing.init_times ();
    Timing.time 1 label run;
    if cfg.time then Timing.print_times ();
  end
      

(* main *)
let _ =
  try
    let cfg = Filt_config.get () in
    if cfg.Filt_config.benchmarks then
      for i = 1 to 30 do
        let size=i*10 in
        let in_img = ref (Ppm.create size size 255 Ppm.black) in
        let out_img = ref (Ppm.create size size 255 Ppm.black) in
        runfilter cfg in_img out_img;
      done
    else begin
      let in_img = ref (Ppm.input_ppm cfg.ic) in
      let out_img = ref (Ppm.init 
                         (!in_img).height 
                         (!in_img).width
                         (!in_img).maxval
                         (Ppm.get_point !in_img)) in
      exec_times cfg.timings_iter 
        (fun () -> runfilter cfg in_img out_img);
      Ppm.output_ppm cfg.oc !in_img;
    end;
    finalize cfg;
  with Defs.Type_error msg->
    begin
      prerr_string "Type error: ";
      prerr_string msg;
      prerr_string "\n";
      exit 1;
    end
  | Parsing.Parse_error ->
    begin
      exit 1;
    end


