(*

  application: image filtering
  state: experimental
  revision: $Id: ppm.ml,v 1.9 2004/11/20 14:25:35 tobias Exp $
 
  Metaprogramming Project
  Authors: Christoph A. Herrmann and Tobias Langhammer
  Chair of Programming, 2004
  University of Passau

 Copyright (C) 2004    University of Passau, Department of Informatics
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*)

open Defs 

(****************** point type and respective members *****************)

type point = { redChannel:int; greenChannel:int; blueChannel:int }

let black = { redChannel=0; greenChannel=0; blueChannel=0 }

let gray_of c = { redChannel = c; greenChannel = c; blueChannel = c}

let point_of_array a =  { redChannel = a.(0); greenChannel = a.(1); blueChannel = a.(2)}
let string_of_point px = 
  begin
    let bf = Buffer.create 40 in
    Buffer.add_string bf "{ red=";
    Buffer.add_string bf (string_of_int px.redChannel);
    Buffer.add_string bf " green=";
    Buffer.add_string bf (string_of_int px.greenChannel);
    Buffer.add_string bf " blue=";
    Buffer.add_string bf (string_of_int px.blueChannel);
    Buffer.add_string bf " }";
    Buffer.contents bf
  end

(***************** image type and respective members ****************)
type ppm_image = 
  {
    height: int; 
    width: int ; 
    maxval: int; 
    raster: (point array) array }

let create height width maxval pix =
  { width=width;
    height=height; 
    maxval=maxval;
    raster = Array.create_matrix height width pix
  }

let init_matrix height width ini =
  let inirow i = Array.init width (ini i) in
  Array.init height inirow

let init height width maxval ini =
  { width=width;
    height=height; 
    maxval=maxval;
    raster = init_matrix height width ini
  }
  

    
  
let get_dims img =(img.height, img.width, img.maxval) 

let string_of_ppm_image img = 
  begin
    let bf = Buffer.create 40 in
    Buffer.add_string bf "{ width=";
    Buffer.add_string bf (string_of_int img.width);
    Buffer.add_string bf " height=";
    Buffer.add_string bf (string_of_int img.height);
    Buffer.add_string bf " maxval=";
    Buffer.add_string bf (string_of_int img.maxval);
    Buffer.add_string bf " raster=<<END\n";
    Array.iter (Array.iter (fun v -> Buffer.add_string bf (string_of_point v)))
       img.raster;
    Buffer.add_string bf "\nEND\n}\n";
    Buffer.contents bf
  end

let get_point img row col = img.raster.(row).(col) 
let get_point_red img row col = img.raster.(row).(col).redChannel 
let get_point_green img row col = img.raster.(row).(col).greenChannel
let get_point_blue img row col = img.raster.(row).(col).blueChannel
let set_point img row col pix = img.raster.(row).(col) <- pix 

exception UnexpectedInput of string

(******************* reading PPM images *********************************)
(* functions for testing membership of chars in char classes *)
let is_member c cs = List.exists (fun n -> n=c) cs

let is_num c = is_member c ['0';'1';'2';'3';'4';'5';'6';'7';'8';'9'] 
let is_whitespace c = is_member c [' ';'\t';'\n'] 


(* functions for reading ppm image from in_channel *)
let current_char = ref '\000' ;;

let skip_comments ic = 
  begin 
    while !current_char <> '\n' do
      current_char := input_char ic
    done;
    current_char := input_char ic
  end

let skip_whitespaces ic = 
  while is_whitespace !current_char do
     current_char := input_char ic
  done

(* parses next int from in_channel while 
 * skipping comments and whitespaces 
 *)
let rec input_int ic =
  if !current_char = '#' then 
    (skip_comments ic; input_int ic)
  else if is_whitespace !current_char then 
    (skip_whitespaces ic; input_int ic)
  else if is_num !current_char then
    begin
      let bf = Buffer.create 10 in
      while is_num !current_char do
        Buffer.add_char bf !current_char;
        current_char := input_char ic
      done;
      int_of_string (Buffer.contents bf)
    end
  else raise (UnexpectedInput (String.make 1 !current_char));
;;


let read_ppm_head ic =
  let magic = "  " in
  let nr = input ic magic 0 2 in
  if nr <> 2 || magic <> "P6"then
    raise (UnexpectedInput ("False magic number: \""^magic^"\"<>\"P6\""));
  current_char := input_char ic;
  let width  = input_int ic in
  let height = input_int ic in
  { width  = width;
    height = height;
    maxval = input_int ic;
    raster = Array.create_matrix height width black
  }

(* reads image raster raster and writes it to img.raster*)
let read_raster ic img =
  let read_int() = 
    let byte = input_byte ic in
    if img.maxval < 256 then 
      byte 
    else 
      byte * 256 + input_byte ic
  in
  let read_point _ = 
    let red = read_int () in
    let green = read_int () in 
    let blue  = read_int () in
    { redChannel = red; 
      greenChannel = green; 
      blueChannel  = blue; } in
  for row = 0 to img.height-1 do
    for col = 0 to img.width-1 do
      img.raster.(row).(col) <- read_point ();
    done
  done

let input_ppm ic =
  begin
    let img=read_ppm_head ic in
    if not (is_whitespace !current_char) then
      raise (UnexpectedInput (String.make 1 !current_char));
    read_raster ic img;
    img
  end

let load_ppm_file filename =
  begin
    let ic = open_in filename in
    let img=input_ppm ic in
    close_in ic;
    img
  end

(************* writing PPM images *****************************)

let output_ppm oc img = 
  begin
    output_string oc "P6\n";
    output_string oc (
    "# generated by PPM.ml of the\n" ^
    "# Metaprogramming Project\n" ^
    "# Authors: Christoph A. Herrmann and Tobias Langhammer\n" ^
    "# University of Passau, Department of Informatics, Chair of Programming\n"^
    "# \n");    
    output_string oc (string_of_int img.width);
    output_string oc " ";
    output_string oc (string_of_int img.height);
    output_string oc "\n";
    output_string oc (string_of_int img.maxval);
    output_string oc "\n";
    let write_int px =
      if img.maxval >= 256 then
        output_byte oc (px / 256);
      output_byte oc px 
    in
    let write_point px =
      write_int px.redChannel;
      write_int px.greenChannel;
      write_int px.blueChannel;
    in
    Array.iter (Array.iter write_point) img.raster
  end

let store_ppm_file filename img =
  begin
    let oc = open_out filename in
    output_ppm oc img;
    close_out oc
  end
(*
let _ = output_ppm stdout (input_ppm stdin)
*)

