(*

  application: image filtering
  state: experimental
  revision: $Id: staticeval.ml,v 1.12 2004/11/20 14:25:35 tobias Exp $
 
  Metaprogramming Project
  Authors: Christoph A. Herrmann and Tobias Langhammer
  University of Passau, Department of Informatics, Chair of Programming

 Copyright (C) 2004    University of Passau, Department of Informatics
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*)


open Defs
open Ppm

let unopfun op a =
  match (op,a) with
    (NegI,VInt a) -> VInt (0 - a)
  | (NegF,VFloat a) -> VFloat (0.0 -. a) 
  | (Not,VBool a) -> VBool (not a)
  | (AbsI, VInt a) -> VInt (abs a)
  | (AbsF, VFloat a) -> VFloat (if a<0.0 then -. a else a)
  | (Exp, VFloat a) -> VFloat (exp a)
  | (Log, VFloat a) -> VFloat (log a)
  | (Sqrt, VFloat a) -> VFloat (sqrt a)
  | (Sin, VFloat a) -> VFloat (sin a)
  | (Cos, VFloat a) -> VFloat (cos a)
  | (Tan, VFloat a) -> VFloat (tan a)
  | (ASin, VFloat a) -> VFloat (asin a)
  | (ACos, VFloat a) -> VFloat (acos a)
  | (ATan, VFloat a) -> VFloat (atan a)
  | _ -> raise (Type_error "in 'unopfun'")

let binopfun op a b = 
    match (op,a,b) with
	(AddI, VInt a, VInt b) -> VInt (a+b)
      | (SubI, VInt a, VInt b) -> VInt (a-b)
      | (MulI, VInt a, VInt b) -> VInt (a*b)
      | (DivI, VInt a, VInt b) -> VInt (a/b)
      | (ModI, VInt a, VInt b) -> VInt (a mod b)
      | (AddF, VFloat a, VFloat b) -> VFloat (a+.b)
      | (SubF, VFloat a, VFloat b) -> VFloat (a-.b)
      | (MulF, VFloat a, VFloat b) -> VFloat (a*.b)
      | (DivF, VFloat a, VFloat b) -> VFloat (a/.b)
      | (PowF, VFloat a, VFloat b) -> VFloat (a**b)
      | (EqB, VBool a, VBool b) -> VBool (a=b)
      | (UEqB, VBool a, VBool b) -> VBool (a!=b)
      | (AndB, VBool a, VBool b) -> VBool (a&&b)
      | (OrB, VBool a, VBool b) -> VBool (a||b)
      | (MinI, VInt a, VInt b) -> VInt (min a b)
      | (MaxI, VInt a, VInt b) -> VInt (max a b)
      | (MinF, VFloat a, VFloat b) -> VFloat (min a b)
      | (MaxF, VFloat a, VFloat b) -> VFloat (max a b)
      | (EqI, VInt a, VInt b) -> VBool (a=b)
      | (UEqI, VInt a, VInt b) -> VBool (a!=b)
      | (LtI, VInt a, VInt b) -> VBool (a<b)
      | (LeI, VInt a, VInt b) -> VBool (a<=b)
      | (GtI, VInt a, VInt b) -> VBool (a>b)
      | (GeI, VInt a, VInt b) -> VBool (a>=b)
      | (EqF, VFloat a, VFloat b) -> VBool (a=b)
      | (UEqF, VFloat a, VFloat b) -> VBool (a!=b)
      | (LtF, VFloat a, VFloat b) -> VBool (a<b)
      | (LeF, VFloat a, VFloat b) -> VBool (a<=b)
      | (GtF, VFloat a, VFloat b) -> VBool (a>b)
      | (GeF, VFloat a, VFloat b) -> VBool (a>=b)
      | _ -> raise (Type_error "in 'binopfun'")


              

let unVInt = function
  | VInt i -> i
  | _ -> raise (Type_error "in 'unVInt'")

let unVFloat = function
  | VFloat f -> f
  | _ -> raise (Type_error "in 'unVFloat'")

let exp_of_value = function
  | VBool i -> Node (Bool, C (VBool i),[])
  | VInt i -> Node (Int, C (VInt i),[])
  | VFloat i -> Node (Float, C (VFloat i),[])
  | VMatrix i -> Node (Matrix, C (VMatrix i),[])

(* simplify: exp -> value environment -> exp *)
let rec simplify expr env =
  let Node (dtype,op,args) = expr in
  let subeval x = simplify x env in
  let arec = List.map subeval in
  let exp_of_args args = Node (dtype, op, args) in
  match op with
  | C v -> exp_of_value v
  | V s ->
    begin match lookup s env with
      | Some exp -> exp_of_value exp
      | None -> expr
    end
  | Read color -> exp_of_args (arec args) 
  | Int2Float ->
      let subexp = head (arec args) in
      if unC subexp != None then
        match subexp with
        | Node (Int, C (VInt i), []) ->
            Node (Float, C (VFloat (float_of_int i)),[])
        | _ -> raise (Type_error "**must not happen**")
      else
        exp_of_args [subexp]
  | Floor -> 
      let subexp = head (arec args) in
      if unC subexp != None then
        match subexp with
        | Node (Float, C (VFloat i), []) ->
            Node (Int, C (VInt (int_of_float i)),[])
        | _ -> raise (Type_error "**must not happen**")
      else
        exp_of_args [subexp]
  | UnOp op ->
      let subexp = head (arec args) in
      begin match unC subexp with
      | Some a -> exp_of_value (unopfun op a)
      | None -> exp_of_args [subexp]
      end
  | BinOp op ->
      let subexps = arec args in
      begin match List.map unC subexps with
      | [Some a; Some b] -> exp_of_value (binopfun op a b)
      | _ -> exp_of_args subexps
      end
  | If ->
      let (cond,a,b) = head3 args in
      let cond' = subeval cond in
      begin match unC cond' with
      | Some (VBool c) ->
          if c then subeval a
          else subeval b
      | Some _ -> raise (Type_error "**must not happen**")
      | None -> exp_of_args (cond'::arec [a;b])
      end
  | Let s -> 
      let (rhs,body) = head2 args in
      let rhs' = subeval rhs in
      begin match unC rhs' with
      | Some v -> simplify body (extEnv (s,v) env) 
      | None -> 
          let body' = simplify body env in
          begin match unC body' with
          | Some v -> exp_of_value v
          | None -> exp_of_args [rhs';body']
          end
      end
  | IndexMatrix s ->
      let (x,y) = head2 (arec args) in
      begin match unC x, unC y, lookup s env with
      | Some (VInt x), Some (VInt y), Some (VMatrix m) -> 
          exp_of_value (VFloat (m.(x).(y)))
      | _ -> raise (Type_error "**must not happen**")
      end
  | Sum s ->
     let (lb,ub,body) = head3 args in
     let (lb',ub') = head2 (arec [lb;ub]) in
     begin match unC lb', unC ub' with
     | Some (VInt l), Some (VInt u) -> 
        let iteration i = 
          let env' = extEnv (s,VInt i) env in
          simplify body env' in
        let step exp i =
          let res = iteration i in
          match (unC exp, unC res, dtype) with
          | Some (VInt a), Some (VInt b), Int ->
              exp_of_value (VInt (a + b))
          | Some (VFloat a), Some (VFloat b), Float -> 
              exp_of_value (VFloat (a +. b))
          | _ ,_, Int   -> Node (Int, BinOp AddI, [exp; res])
          | _ ,_, Float -> Node (Int, BinOp AddF, [exp; res])
          | _ -> raise (Type_error
             "in 'simplify': False summation types")
        in
        let head =
          if l>u then begin 
            match dtype with 
            | Int -> exp_of_value (VInt 0)
            | Float -> exp_of_value (VFloat 0.0)
            | _ -> raise (Type_error
               "in 'simplify': False summation types")
          end else iteration l
        in
        List.fold_left step head (fromto (l+1) u)
     | _,_ -> 
        exp_of_args [lb';ub';simplify body env]
     end
      
      
let newenv pairs = List.fold_right extEnv pairs emptyEnv 
     
  
let simplify expr width height iter =
  let env = 
    newenv [("iter",VInt iter);("width",VInt width);("height",VInt height)]
  in
  simplify expr env

let parse funString =
  Filt_parser.main Filt_lexer.token (Lexing.from_string funString)
 


